//#####################################################################
// Copyright 2012.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################

#include <Core/Arrays/ARRAYS_ND.h>
#include <Core/Arrays/FACE_ARRAYS.h>
#include <Core/Boundaries/BOUNDARY_PERIODIC.h>
#include <Core/Boundaries/PERIODIC_BOUNDARY_HELPER.h>
#include <Core/Fourier_Transforms/FFT.h>
#include <Core/Grids/FACE_INDEX.h>
#include <Core/Log/LOG.h>
#include <Core/Matrices/LAPACK.h>
#include <Core/Matrices/MATRIX_MXN.h>
#include "PROJECT_FFT.h"

namespace PhysBAM
{

// #if 0
// int shift=0; // offset shift to i+shift
// #else
// int shift=-1; // offset shift to i+shift
// #endif

template<class TV,class IV,class T,int d>
void Test_Stencils(const ARRAY<T,FACE_INDEX<d>>& mac,const IV& size,const GRID<TV>& grid,
    const ARRAY<CONSTRAINT_STENCIL<TV>>& stencil)
{
    ARRAY<T> max_err(stencil.m);
    
    for(RANGE_ITERATOR<d> it(size);it.Valid();it.Next())
    {
        for(int i=0;i<stencil.m;i++)
        {
            T error=0;
            for(const auto& j:stencil(i).constraints)
            {
                auto f=j.first;
                f.index=wrap(f.index+it.index,{},size);
                error+=j.second*mac(f);
            }
            max_err(i)=max(max_err(i),abs(error));
        }
    }
    LOG::printf("max_err %P\n",max_err);
}

template<class TV,class IV,class T,int d>
void Test_Div(const ARRAY<T,FACE_INDEX<d>>& mac,const IV& size,const GRID<TV>& grid,
    const ARRAY<T>& stencil)
{
    ARRAY<CONSTRAINT_STENCIL<TV>> st;
    Make_Div_Stencil(st,stencil);
    Test_Stencils(mac,size,grid,st);
}

template<class TV,class IV,class T,int d>
void Test_Curl(const ARRAY<T,FACE_INDEX<d>>& mac,const IV& size,const GRID<TV>& grid,
    const ARRAY<T>& stencil)
{
    ARRAY<CONSTRAINT_STENCIL<TV>> st;
    Make_Curl_Stencil(st,stencil);
    //for (auto cs : st) {
        //for (auto p : cs.constraints) {
            //LOG::printf("%p*%p + ", p.second, grid.Face_X(p.first));
        //}
    //}
    Test_Stencils(mac,size,grid,st);
}

template<class T,class TV>
void Make_Div_Stencil(ARRAY<CONSTRAINT_STENCIL<TV>>& out,const ARRAY<T>& in)
{
    typedef VECTOR<int,TV::m> IV;
    int shift=-(in.m-1)/2;
    CONSTRAINT_STENCIL<TV> st;
    for(int i=0;i<TV::m;i++)
        for(int j=0;j<in.m;j++)
        {
            IV k;
            k(i)+=j+shift;
            st.constraints.Append({{i,k},in(j)});
        }
    out.Append(st);
}

template<class T,class TV>
void Make_Curl_Stencil(ARRAY<CONSTRAINT_STENCIL<TV>>& out,const ARRAY<T>& in)
{
    typedef VECTOR<int,TV::m> IV;
    int shift=-(in.m-1)/2;
    for(int a=0;a<TV::m;a++)
        for(int b=a+1;b<TV::m;b++)
        {
            CONSTRAINT_STENCIL<TV> st;
            for(int j=0;j<in.m;j++)
            {
                IV k,l;
                k(a)-=j+shift;
                l(b)-=j+shift;
                st.constraints.Append({{b,k},in(in.m-j-1)});
                st.constraints.Append({{a,l},in(j)});
            }
            out.Append(st);
        }

}

template<class T,int d,class TV>
void Project_Div(const GRID<TV>& grid,ARRAY<T,FACE_INDEX<d>>& mac,
    const ARRAY<T>& stencil,T tol)
{
    ARRAY<CONSTRAINT_STENCIL<TV>> st;
    Make_Div_Stencil(st,stencil);
    Project_Stencils(grid,mac,st,tol);
    return;
}

template<class T,int d,class TV>
void Project_Curl(const GRID<TV>& grid,ARRAY<T,FACE_INDEX<d>>& mac,
    const ARRAY<T>& stencil,T tol)
{
    ARRAY<CONSTRAINT_STENCIL<TV>> st;
    Make_Curl_Stencil(st,stencil);
    Project_Stencils(grid,mac,st,tol);
    return;
}


template<class T,int d,class TV>
void Project_Stencils(const GRID<TV>& grid,ARRAY<T,FACE_INDEX<d>>& mac,
    const ARRAY<CONSTRAINT_STENCIL<TV>>& stencil,T tol)
{
    typedef VECTOR<int,TV::m> IV;
    typedef std::complex<T> C;
    typedef VECTOR<C,TV::m> CV;

    VECTOR<ARRAY<T,IV>,TV::m> u;
    ARRAY<VECTOR<ARRAY<T,IV>,TV::m>> st(stencil.m);

    for(int i=0;i<TV::m;i++)
    {
        u(i).Resize(grid.numbers_of_cells);
        u(i).Limited_Shifted_Get(u(i),mac.Component(i),{});
        u(i).array-=u(i).array.Average();
        for(auto&a:st)
            a(i).Resize(grid.numbers_of_cells);
    }

    for(int j=0;j<stencil.m;j++)
    {
        auto&s=st(j);
        for(const auto& t:stencil(j).constraints)
        {
            auto idx=wrap(t.first.index,IV(),grid.numbers_of_cells);
            s(t.first.axis)(idx)+=t.second;
        }
    }
    
    VECTOR<ARRAY<C,IV>,TV::m> fu;
    ARRAY<VECTOR<ARRAY<C,IV>,TV::m>> fst(stencil.m);

    FFT<TV> fft;
    for(int i=0;i<TV::m;i++)
    {
        fu(i).Resize(grid.numbers_of_cells);
        fft.Transform(u(i),fu(i));
        for(int j=0;j<stencil.m;j++)
        {
            fst(j)(i).Resize(grid.numbers_of_cells);
            fft.Transform(st(j)(i),fst(j)(i));
        }
    }

    for(RANGE_ITERATOR<TV::m> it(grid.numbers_of_cells);it.Valid();it.Next())
    {
        CV uf;
        MATRIX_MXN<C> w(stencil.m,TV::m),cw(stencil.m,TV::m);
        for(int i=0;i<TV::m;i++)
        {
            for(int j=0;j<stencil.m;j++)
            {
                C z=fst(j)(i)(it.index)*grid.one_over_dX(i);
                w(j,i)=z;
                cw(j,i)=conj(z);
            }
            uf(i)=fu(i)(it.index);
        }

        auto M=cw.Times_Transpose(w),MM=M;
        ARRAY<C> rhs=cw*uf,sol(rhs.m);
        Least_Squares_Solve_Raw(M,rhs,sol);
        uf-=CV(w.Transpose_Times(sol));
        for(int i=0;i<TV::m;i++)
            fu(i)(it.index)=uf(i);
    }

    for(int i=0;i<TV::m;i++)
    {
        fft.Inverse_Transform(fu(i),u(i));
        u(i).Limited_Shifted_Get(mac.Component(i),u(i),{});
    }

    BOUNDARY_PERIODIC<TV,T> boundary;
    boundary.Apply_Boundary_Condition_Face(grid,mac);
    Fill_Ghost_Faces_Periodic(grid, mac, VECTOR<bool, TV::m>::All_Ones_Vector());
}





template void Project_Div<float,2,VECTOR<float, 2>>(const GRID<VECTOR<float,2>>& grid,ARRAY<float,FACE_INDEX<2>>& mac,const ARRAY<float>& stencil,float tol);
template void Project_Div<float,3,VECTOR<float, 3>>(const GRID<VECTOR<float,3>>& grid,ARRAY<float,FACE_INDEX<3>>& mac,const ARRAY<float>& stencil,float tol);
template void Project_Div<double,2,VECTOR<double, 2>>(const GRID<VECTOR<double,2>>& grid,ARRAY<double,FACE_INDEX<2>>& mac,const ARRAY<double>& stencil,double tol);
template void Project_Div<double,3,VECTOR<double, 3>>(const GRID<VECTOR<double,3>>& grid,ARRAY<double,FACE_INDEX<3>>& mac,const ARRAY<double>& stencil,double tol);

template void Test_Div(const ARRAY<float,FACE_INDEX<2>>& mac,const VECTOR<int,2>& size,const GRID<VECTOR<float,2>>& grid, const ARRAY<float>& stencil);
template void Test_Div(const ARRAY<float,FACE_INDEX<3>>& mac,const VECTOR<int,3>& size,const GRID<VECTOR<float,3>>& grid, const ARRAY<float>& stencil);
template void Test_Div(const ARRAY<double,FACE_INDEX<2>>& mac,const VECTOR<int,2>& size,const GRID<VECTOR<double,2>>& grid, const ARRAY<double>& stencil);
template void Test_Div(const ARRAY<double,FACE_INDEX<3>>& mac,const VECTOR<int,3>& size,const GRID<VECTOR<double,3>>& grid, const ARRAY<double>& stencil);

template void Project_Stencils(const GRID<VECTOR<float,3>>& grid,ARRAY<float,FACE_INDEX<3>>& mac,const ARRAY<CONSTRAINT_STENCIL<VECTOR<float,3>>>& stencil,float tol);
template void Test_Stencils<VECTOR<double,2>,VECTOR<int,2>,double,2>(ARRAY<double,FACE_INDEX<2> > const&,VECTOR<int,2> const&,GRID<VECTOR<double,2> > const&,ARRAY<CONSTRAINT_STENCIL<VECTOR<double,2> >,int> const&);
template void Test_Stencils<VECTOR<double,3>,VECTOR<int,3>,double,3>(ARRAY<double,FACE_INDEX<3> > const&,VECTOR<int,3> const&,GRID<VECTOR<double,3> > const&,ARRAY<CONSTRAINT_STENCIL<VECTOR<double,3> >,int> const&);
template void Test_Stencils<VECTOR<float,2>,VECTOR<int,2>,float,2>(ARRAY<float,FACE_INDEX<2> > const&,VECTOR<int,2> const&,GRID<VECTOR<float,2> > const&,ARRAY<CONSTRAINT_STENCIL<VECTOR<float,2> >,int> const&);
template void Test_Stencils<VECTOR<float,3>,VECTOR<int,3>,float,3>(ARRAY<float,FACE_INDEX<3> > const&,VECTOR<int,3> const&,GRID<VECTOR<float,3> > const&,ARRAY<CONSTRAINT_STENCIL<VECTOR<float,3> >,int> const&);
template void Project_Curl<double,3,VECTOR<double,3> >(GRID<VECTOR<double,3> > const&,ARRAY<double,FACE_INDEX<3> >&,ARRAY<double,int> const&,double);
template void Test_Curl<VECTOR<double,3>,VECTOR<int,3>,double,3>(ARRAY<double,FACE_INDEX<3> > const&,VECTOR<int,3> const&,GRID<VECTOR<double,3> > const&,ARRAY<double,int> const&);
template void Project_Curl<double,2,VECTOR<double,2> >(GRID<VECTOR<double,2> > const&,ARRAY<double,FACE_INDEX<2> >&,ARRAY<double,int> const&,double);
template void Test_Curl<VECTOR<double,2>,VECTOR<int,2>,double,2>(ARRAY<double,FACE_INDEX<2> > const&,VECTOR<int,2> const&,GRID<VECTOR<double,2> > const&,ARRAY<double,int> const&);
}
