//#####################################################################
// Copyright 2021, Craig Schroeder, Tamar Shinar.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Arrays/ARRAYS_ND.h>
#include <Core/Arrays/FACE_ARRAYS.h>
#include <Core/Boundaries/BOUNDARY.h>
#include <Core/Geometry_Particles/DEBUG_PARTICLES.h>
#include <Core/Grids/CELL_ITERATOR.h>
#include <Core/Grids/FACE_INDEX.h>
#include <Core/Grids/FACE_RANGE_ITERATOR.h>
#include <Core/Grids/GRID.h>
#include <Core/Interpolation/CUBIC_INTERPOLATION_FACE.h>
#include <Core/Interpolation/CUBIC_SPLINE_INTERPOLATION_FACE.h>
#include <Core/Interpolation/DIV_FREE_BALSARA.h>
#include <Core/Interpolation/DIV_FREE_BALSARA_WENO.h>
#include <Core/Interpolation/DIV_FREE_INTERPOLATION.h>
#include <Core/Interpolation/DIV_FREE_INTERPOLATION_C1.h>
#include <Core/Interpolation/DIV_FREE_INTERPOLATION_C1_INTERP.h>
#include <Core/Interpolation/DIV_FREE_INTERPOLATION_INTERP.h>
#include <Core/Interpolation/GENERAL_INTERPOLATION_FACE.h>
#include <Core/Interpolation/LINEAR_INTERPOLATION_FACE.h>
#include <Core/Parsing/PARSE_ARGS.h>
#include <Core/Random_Numbers/RANDOM_NUMBERS.h>
#include <Core/Seeding/POISSON_DISK.h>
#include <Core/Utilities/VIEWER_DIR.h>
#include <Core/Vectors/VECTOR.h>
#include <Core/Viewer/VIEWER_OUTPUT.h>
#include "INTERPOLATION_SCHEMES.h"
#include "VELOCITY_FIELDS.h"
#include <fstream>

using namespace PhysBAM;

typedef float T;

template<int d>
void Run(PARSE_ARGS& parse_args)
{
    typedef VECTOR<T,d> TV;
    typedef VECTOR<int,TV::m> IV;
    std::string sim_dir="path-grid-out";
    int velocity_field=0;
    bool plot_error=false;
    parse_args.Add("-sim",&sim_dir,"dir","simulation directory to trace particles on");
    parse_args.Add("-vel",&velocity_field,"velocity field","which velocity field to use (0-3)");
    parse_args.Add("-err",&plot_error,"plot errors instead of velocities");
    parse_args.Parse();
    
    VIEWER_DIR viewer_dir(sim_dir);
    STREAM_TYPE stream_type=STREAM_TYPE::type_float;
    VIEWER_OUTPUT viewer_output(stream_type,viewer_dir);
    GRID<TV> grid(IV()+8,RANGE<TV>::Unit_Box());
    if(d==2) viewer_output.Add_Global({.is_2d=true});
    if(d==3) viewer_output.template Add_Axes<TV::m>("axes",{});
    viewer_output.Add("debug_particles",DEBUG_PARTICLES<TV>::Get(),false,{});
    viewer_output.Add("grid",grid,true,{.color={{1,0,0}}});

    int ghost=8;

    ARRAY<T,FACE_INDEX<TV::m> > data(grid.Cell_Indices(ghost));
    ARRAY<T,FACE_INDEX<TV::m> > data_avg(grid.Cell_Indices(ghost));
    IV Ai,Bi;
    Ai(0)=1;
    Ai(1)=2;
    if constexpr(TV::m==3) Ai(2)=1;
    Bi(0)=7;
    Bi(1)=4;
    if constexpr(TV::m==3) Bi(2)=5;

    TV A=grid.Face_X({0,Ai}),B=grid.Face_X({0,Bi});
    T l=(B-A).Magnitude();

    Add_Point(A,{1,0,0});
    Add_Point(B,{1,0,0});
    Add_Segment(A,B,{1,.5,0});

    std::ofstream cross_u("path-grid-cross-u.txt");
    std::ofstream cross_v("path-grid-cross-v.txt");
    LOG::fprintf(cross_u,"t u\n");
    LOG::fprintf(cross_v,"t v\n");
    auto field=Velocity_Field<TV>(velocity_field);
    auto field_avg=Velocity_Field_Average<TV>(velocity_field);

    for(FACE_RANGE_ITERATOR<TV::m> it(data.domain);it.Valid();it.Next())
    {
        TV X=grid.Face_X(it.face);
        data(it.face)=field(X)(it.face.axis);
        data_avg(it.face)=field_avg(X, grid.dX.Max())(it.face.axis);
        T a=(X-A).Magnitude(),b=(X-B).Magnitude();
        if(abs(a+b-l)<1e-6)
            LOG::fprintf(it.face.axis?cross_v:cross_u,"%P %P\n",a/l,plot_error?0:data(it.face));
    }

    std::ofstream fout("path-grid-data.txt");
    auto interps=Interpolation_Schemes<TV>();
    LOG::fprintf(fout,"t");
    for(auto s:interps)
    {
        if(TV::m==2) LOG::fprintf(fout," %s_u %s_v",s.name,s.name);
        if(TV::m==3) LOG::fprintf(fout," %s_u %s_v %s_w",s.name,s.name,s.name);
    }
    LOG::fprintf(fout,"\n");
    int n=1000;

    IV prev_cell = grid.Cell_I(A);

    for(int i=0;i<=n;i++)
    {
        T t=(T)i/n;
        TV X=A+t*(B-A);
        IV cell=grid.Cell_I(X);

        if(cell.x != prev_cell.x)
        {
            LOG::printf("u %P\n", t);
        }

        if(cell.y != prev_cell.y)
        {
            LOG::printf("v %P\n", t);
        }
        prev_cell = cell;

        TV base;
        if(plot_error) base=Velocity_Field<TV>(velocity_field)(X);
        LOG::fprintf(fout,"%P",t);
        for(auto s:interps) {
            if (s.need_avg) {
                LOG::fprintf(fout," %P",s.interp->Interpolate_Face(grid,data_avg,X)-base);
            } else {
                LOG::fprintf(fout," %P",s.interp->Interpolate_Face(grid,data,X)-base);
            }
        }
        LOG::fprintf(fout,"\n");
    }

    Flush_Frame();
}

int main(int argc, char* argv[])
{
    PARSE_ARGS parse_args(argc,argv);
    int dim=2;
    parse_args.Add("-dim",&dim,"dim","dimension to test");
    parse_args.Parse(true);
    if(dim==2) Run<2>(parse_args);
    else Run<3>(parse_args);
    
    return 0;
}

