#include <Core/Arrays/ARRAY.h>
#include <Core/Grids/FACE_INDEX.h>
#include <Core/Grids/FACE_ITERATOR.h>
#include <Core/Grids/GRID.h>
#include <Core/Interpolation/CUBIC_INTERPOLATION_FACE.h>
#include <Core/Interpolation/CUBIC_SPLINE_INTERPOLATION_FACE.h>
#include <Core/Interpolation/DIV_FREE_BALSARA.h>
#include <Core/Interpolation/DIV_FREE_INTERPOLATION.h>
#include <Core/Interpolation/DIV_FREE_INTERPOLATION_C1.h>
#include <Core/Interpolation/DIV_FREE_INTERPOLATION_C1_INTERP.h>
#include <Core/Interpolation/DIV_FREE_INTERPOLATION_INTERP.h>
#include <Core/Interpolation/GENERAL_INTERPOLATION_FACE.h>
#include <Core/Interpolation/LINEAR_INTERPOLATION_FACE.h>
#include <Core/Parsing/PARSE_ARGS.h>
#include <Core/Random_Numbers/RANDOM_NUMBERS.h>
#include <Core/Vectors/VECTOR.h>
#include "VELOCITY_FIELDS.h"

using namespace PhysBAM;

typedef double RW;
typedef double T;

template<class TV>
void Test_Resolution(int res,const ARRAY<TV>& positions,
    const ARRAY<TV>& exact_velocities,int velocity_field)
{
    typedef VECTOR<int,TV::m> IV;
    GRID<TV> grid(IV()+res,RANGE<TV>::Unit_Box());

    ARRAY<T,FACE_INDEX<TV::m>> velocity(grid.Cell_Indices());
    auto field=Velocity_Field<TV>(velocity_field);
    for(FACE_ITERATOR it(grid); it.Valid(); it.Next()) 
    {
        auto fi=it.Full_Index();
        TV X=it.Location();

        velocity(fi)=field(X)[fi.axis];
    }

    LINEAR_INTERPOLATION_FACE<TV,T> i1;
    CUBIC_INTERPOLATION_FACE<TV,T> i2;
    CUBIC_SPLINE_INTERPOLATION_FACE<TV,T> i3;
    DIV_FREE_BALSARA<TV,T> i4;
    DIV_FREE_INTERPOLATION<TV,T> i5;
    DIV_FREE_INTERPOLATION_INTERP<TV,T> i6;
    DIV_FREE_INTERPOLATION_C1<TV,T> i7;
    DIV_FREE_INTERPOLATION_C1_INTERP<TV,T> i8;

    INTERPOLATION_FACE<TV,T>* interps[]={&i1,&i2,&i3,&i4,&i5,&i6,&i7,&i8};

    LOG::printf("%P",res);
    for(int i=0; i<8; i++) 
    {
        INTERPOLATION_FACE<TV,T>* interp=interps[i];
        T max_err=0;;

        for(int p=0; p<positions.m; p++) 
        {
            TV V=interp->Interpolate_Face(grid,velocity,positions(p));
            T err=(V - exact_velocities(p)).Magnitude();
            max_err=max(max_err,err);
        }

        LOG::printf(" %P",max_err);
    }
    LOG::printf("\n");
}

template<int d>
void Run(PARSE_ARGS& parse_args)
{
    typedef VECTOR<T,d> TV;
    typedef VECTOR<int,TV::m> IV;
    int velocity_field=1;
    int refine=8;
    parse_args.Add("-vel",&velocity_field,"velocity field","which velocity field to use (0-3)");
    parse_args.Add("-refine",&refine,"refinement","refinement level");
    parse_args.Parse();

    RANDOM_NUMBERS<T> rand(123);

    ARRAY<TV> positions(1000000);
    rand.Fill_Uniform(positions,TV(),TV()+1);

    ARRAY<TV> velocities(positions.m);
    auto field=Velocity_Field<TV>(velocity_field);
    for(int p=0; p<positions.m; p++) 
    {
        velocities(p)=field(positions(p));
    }

    Test_Resolution(refine,positions,velocities,velocity_field);
}

int main(int argc,char* argv[])
{
    PARSE_ARGS parse_args(argc,argv);
    int dim=2;
    bool names_only=false;
    parse_args.Add("-dim",&dim,"dim","dimension to test");
    parse_args.Add("-names",&names_only,"print names and exit");
    parse_args.Parse(true);

    if(names_only)
    {
        LOG::printf("res linear cubic cubic-spline balsara div-free-c0 div-free-c0-interp div-free-c1 div-free-c1-interp\n");
        return 0;
    }

    if(dim==2) Run<2>(parse_args);
    else Run<3>(parse_args);
    return 0;
}
