#include <Core/Arrays/ARRAY.h>
#include <Core/Arrays/ARRAYS_ND_VIEW.h>
#include <Core/Basic_Geometry/SEGMENT_2D.h>
#include <Core/Geometry_Particles/DEBUG_PARTICLES.h>
#include <Core/Geometry_Particles/GEOMETRY_PARTICLES.h>
#include <Core/Grids/CELL_ITERATOR.h>
#include <Core/Grids/FACE_INDEX.h>
#include <Core/Grids/FACE_ITERATOR.h>
#include <Core/Grids/GRID.h>
#include <Core/Images/PNG_FILE.h>
#include <Core/Interpolation/CUBIC_INTERPOLATION_FACE.h>
#include <Core/Interpolation/CUBIC_SPLINE_INTERPOLATION_FACE.h>
#include <Core/Interpolation/DIV_FREE_BALSARA.h>
#include <Core/Interpolation/DIV_FREE_BALSARA_WENO.h>
#include <Core/Interpolation/DIV_FREE_INTERPOLATION.h>
#include <Core/Interpolation/DIV_FREE_INTERPOLATION_4TH_ORDER.h>
#include <Core/Interpolation/DIV_FREE_INTERPOLATION_C1.h>
#include <Core/Interpolation/DIV_FREE_INTERPOLATION_C1_INTERP.h>
#include <Core/Interpolation/DIV_FREE_INTERPOLATION_FLUXES.h>
#include <Core/Interpolation/DIV_FREE_INTERPOLATION_INTERP.h>
#include <Core/Interpolation/GENERAL_INTERPOLATION_FACE.h>
#include <Core/Interpolation/LINEAR_INTERPOLATION_FACE.h>
#include <Core/Ordinary_Differential_Equations/RUNGEKUTTA.h>
#include <Core/Parsing/PARSE_ARGS.h>
#include <Core/Random_Numbers/RANDOM_NUMBERS.h>
#include <Core/Read_Write/FILE_UTILITIES.h>
#include <Core/Seeding/POISSON_DISK.h>
#include <Core/Utilities/VIEWER_DIR.h>
#include <Core/Vectors/VECTOR.h>
#include <Core/Viewer/VIEWER_OUTPUT.h>
#include "PROJECT_FFT.h"
#include "VELOCITY_FIELDS.h"

using namespace PhysBAM;

typedef double RW;
typedef double T;
typedef VECTOR<T,2> TV;
typedef VECTOR<int,TV::m> IV;
typedef ARRAY<T,FACE_INDEX<TV::m> > T_FACE_ARRAY;
typedef INTERPOLATION_FACE<TV,T,T_FACE_ARRAY> T_INTERPOLATION_FACE;
typedef ARRAY<VECTOR<double,3>,VECTOR<int,2> > PNG_ARRAY;

//  (continuity)
//  (streamlines: spiral vs. cycle)
//  (stencil: with 4th order Poisson vs 2nd order Poisson)

// BALSARA, LINEAR, DIV FREE 2 (2), DIV FREE 4 (4), DIV FREE 4 (2) , BALSARA 3 vs.
// full domain, 1 streamline


void Seed_Particles(GRID<TV> grid,int ppc,GEOMETRY_PARTICLES<TV>& particles)
{
    POISSON_DISK<TV> poisson_disk(0);
    poisson_disk.Set_Distance_By_Volume(grid.Cell_Size()/(ppc));
    RANDOM_NUMBERS<T> rand(12345);
    ARRAY<TV> X;
    poisson_disk.Sample(rand,grid.domain,X);
    auto* valid=particles.Get_Array<bool>("valid");

    for(auto& x:X)
    {
        int p=particles.Add_Element();
        particles.X(p)=x;
    }
    valid->Fill(true);
}

void Generate_Projected_Field(GRID<TV>& grid, T_FACE_ARRAY& vel_2, T_FACE_ARRAY& vel_4,int seed)
{
    int num_ghost_cells=3;

    ARRAY<T> ST2={-1,1};
    ARRAY<T> ST4={1/24.,-27/24.,27/24.,-1/24.};

    vel_2.Resize(grid.Cell_Indices(num_ghost_cells));
    vel_4.Resize(grid.Cell_Indices(num_ghost_cells));

    RANDOM_NUMBERS<T> rand(seed);
    for(FACE_ITERATOR<TV> it(grid);it.Valid();it.Next())
        vel_4(it.Full_Index()) = vel_2(it.Full_Index()) = rand.Get_Number();

    Project_Div<T,TV::m,TV>(grid,vel_2, ST2);
    Project_Div<T,TV::m,TV>(grid,vel_4, ST4);

    Test_Div(vel_2,grid.numbers_of_cells,grid,ST2);
    Test_Div(vel_4,grid.numbers_of_cells,grid,ST4);
}

TV Advance_Particle(const GRID<TV>& grid,TV X_initial,int rk_order,T dt,const T_FACE_ARRAY& face_velocities,T direction,const T_INTERPOLATION_FACE& interp,bool& valid,VECTOR<T,3> color,const GRID<VECTOR<T,2> >& png_grid,PNG_ARRAY& png_array,VECTOR<T,3> png_color)
{
    /*
    TV v=interp.Interpolate_Face(grid,face_velocities,X_initial)*direction;
    TV X_new=X_initial+.5*dt*v;
    v=interp.Interpolate_Face(grid,face_velocities,X_new)*direction;
    X_new=X_initial+dt*v;
    */          

    TV X_new=X_initial;
    for(RUNGEKUTTA<TV> rk(X_new,rk_order,dt,0);rk.Valid();rk.Next())
    {
        TV v=interp.Interpolate_Face(grid,face_velocities,X_new);
        X_new += dt*v*direction;
    }

    valid = grid.domain.Lazy_Inside(X_new);
    if(!valid) Add_Ball(X_new,.1,color);

    Add_Segment(X_initial,X_new,color);

    // write segment to png array
    int linewidth=3; // linewidth in 1/1000th of image
    IV index_new=png_grid.Cell_I(X_new);
    int cell_radius=png_grid.numbers_of_cells(0)*linewidth/2000;
    RANGE<IV> range(index_new-cell_radius,index_new+cell_radius);
    for(CELL_ITERATOR<TV> it(png_grid,range);it.Valid();it.Next())
        if(png_grid.Valid_Cell(it.index)) png_array(it.index)=png_color;

    return X_new;
}

bool Valid_Streamline(int p,const GRID<TV>& grid,const GEOMETRY_PARTICLES<TV>& particles,const ARRAY<bool,IV>& occupied_cells)
{
    IV index=grid.Cell_I(particles.X(p));
    if(occupied_cells(index)) return false;
    for(int i=0;i<grid.number_of_one_ring_neighbors_per_cell;i++)
        if(occupied_cells(grid.One_Ring_Neighbor(index,i)))
            return false;
    return true;
}

void Write_Png(const std::string& filename,const PNG_ARRAY& png_array,int antialiasing,int resolution,const GRID<TV>& png_grid,TV starting_pt=TV())
{
    if(antialiasing<=1)
    {
        PNG_FILE<T>::Write(filename,png_array);
        return;
    }

    GRID<VECTOR<T,2> > png_grid_antialiased(VECTOR<int,2>()+resolution,png_grid.domain);
    PNG_ARRAY png_array_antialiased(png_grid_antialiased.Cell_Indices());
    for(CELL_ITERATOR<TV> it(png_grid);it.Valid();it.Next())
        png_array_antialiased(it.index/antialiasing) += png_array(it.index);
    png_array_antialiased.array *= sqr(1./antialiasing);
    if(starting_pt!=TV())
    {
        VECTOR<T,3> starting_color(1,0,0);
        int linewidth=20;
        IV starting_index=png_grid_antialiased.Cell_I(starting_pt);
        int cell_radius=png_grid_antialiased.numbers_of_cells(0)*linewidth/2000;
        RANGE<IV> range(starting_index-cell_radius,starting_index+cell_radius);
        for(CELL_ITERATOR<TV> it(png_grid_antialiased,range);it.Valid();it.Next())
            if(png_grid_antialiased.Valid_Cell(it.index))
                png_array_antialiased(it.index)=starting_color;
    }
    PNG_FILE<T>::Write(filename,png_array_antialiased);
    LOG::cout<<"\t"<<filename<<std::endl;
}

void Run(PARSE_ARGS& parse_args)
{
    int refine=8;
    int spc=1;
    std::string output_dir="streamlines-out";
    int substeps=1000;
    T end_time=2*pi;
    int single_streamline=-1;
    bool debug_streamlines=false;
    int seed=123;

    parse_args.Add("-refine",&refine,"refinement","refinement level");
    parse_args.Add("-spc",&spc,"spc","approximate streamlines per cell");
    parse_args.Add("-o",&output_dir,"output directory","output directory");
    parse_args.Add("-substeps",&substeps,"substeps","number of substeps");
    parse_args.Add("-end_time",&end_time,"end time","end time");
    parse_args.Add("-s",&single_streamline,"p","streamline to draw");
    parse_args.Add("-seed",&seed,"seed","random number seed for velocity field");
    parse_args.Add("-debug",&debug_streamlines,"output debugging information");
    parse_args.Parse(true);

    // create the grid
    RANGE<TV> range=RANGE<TV>::Unit_Box();
    GRID<TV> grid(IV()+refine,range);
    T threshold=grid.dX(0)*1e-3;

    // set up viewer output
    VIEWER_DIR viewer_dir(output_dir);
    STREAM_TYPE stream_type=STREAM_TYPE::type_float;
    VIEWER_OUTPUT viewer_output(stream_type,viewer_dir);
    viewer_output.Add("debug_particles",DEBUG_PARTICLES<TV>::Get(),false,{});
    viewer_output.Add("grid",grid,true,{.color={{1,0,0}}});
    Flush_Frame();
    
    viewer_dir.Set(0);

    // seed the particles on the grid
    GEOMETRY_PARTICLES<TV> particles;
    particles.Add_Array<bool>("valid");
    auto* valid=particles.Get_Array<bool>("valid");
    Seed_Particles(grid,4*spc,particles);
    Write_To_File(stream_type,viewer_dir.Frame_File("mpm_particles"),particles);

    // set up velocity field on the grid
    T_FACE_ARRAY face_velocities_st2, face_velocities_st4;
    Generate_Projected_Field(grid,face_velocities_st2,face_velocities_st4,seed);

    Write_To_File(stream_type,viewer_dir.Frame_File("mac_velocities"),face_velocities_st2);
    LOG::printf("Writing output to %P\n",output_dir);

    DIV_FREE_BALSARA<TV,T> df_b;
    LINEAR_INTERPOLATION_FACE<TV,T> int_l;
    DIV_FREE_INTERPOLATION<TV,T> df_c0;
    DIV_FREE_INTERPOLATION_4TH_ORDER<TV,T> df_4;
//    DIV_FREE_BALSARA_WENO<TV,T,3> df_b_weno;
    CUBIC_SPLINE_INTERPOLATION_FACE<TV,T> i3;
    DIV_FREE_INTERPOLATION_FLUXES<TV,T> i10;

    typedef std::tuple<T_INTERPOLATION_FACE*,std::string,T_FACE_ARRAY*> T_INTERP_TEST;
    ARRAY<T_INTERP_TEST> name = {
        T_INTERP_TEST(&df_c0,"DIV_FREE_INTERPOLATION_2ND_ORDER_ST2",&face_velocities_st2),
        T_INTERP_TEST(&i10,"DIV_FREE_INTERPOLATION_FLUXES",&face_velocities_st2),
        T_INTERP_TEST(&df_b,"DIV_FREE_BALSARA",&face_velocities_st2),
        T_INTERP_TEST(&int_l,"LINEAR_INTERPOLATION_FACE",&face_velocities_st2),
        T_INTERP_TEST(&df_4,"DIV_FREE_INTERPOLATION_4TH_ORDER_ST4",&face_velocities_st4),
        T_INTERP_TEST(&df_4,"DIV_FREE_INTERPOLATION_4TH_ORDER_ST2",&face_velocities_st2),
//        T_INTERP_TEST(&df_b_weno,"DIV_FREE_BALSARA_WENO",&face_velocities_st2),
        T_INTERP_TEST(&i3,"CUBIC_SPLINE_INTERPOLATION_FACE",&face_velocities_st2)
    };

    T dt=end_time/substeps;
    ARRAY<bool> valid_forward(particles.X.m);
    ARRAY<bool> valid_backward(particles.X.m);
    int min_substeps=5;
    VECTOR<T,3> red(1,0,0), cyan(0,1,1), white(1,1,1), green(0,1,0);

    GRID<TV> ogrid((IV()+refine)*spc*2,grid.domain);
    ARRAY<bool,IV> occupied_cells(ogrid.Cell_Indices(2*spc));
    
    int start_streamline=0,end_streamline=particles.X.m-1;
    if(single_streamline>=0) start_streamline=end_streamline=single_streamline;

    int antialiasing=4;
    int resolution=1000;
    GRID<VECTOR<T,2> > png_grid(VECTOR<int,2>()+resolution*antialiasing,grid.domain);
    PNG_ARRAY png_array(png_grid.Cell_Indices());
    ARRAY<VECTOR<T,3>> colors(particles.X.m);
    RANDOM_NUMBERS<T> rand_color(seed);
    rand_color.Fill_Uniform(colors,0.,.75);
    valid->Fill(true);

    LOG::printf("Writing png files:\n");

    for(int fs=0;fs<name.Size();fs++)
    {
        ARRAY<TV> X_forward(particles.X);
        ARRAY<TV> X_backward(particles.X);
        valid_forward.Fill(true);
        valid_backward.Fill(true);
        occupied_cells.Fill(false);
        png_array.Fill(VECTOR<double,3>(1,1,1));

        // integrate the particles forward and backward and create segmented curve
        for(int p=start_streamline;p<=end_streamline;p++)
        {
            if(!(*valid)(p)) continue;
            
            // skip if the starting point is close to any previously drawn streamlines
            if(!fs && !Valid_Streamline(p,ogrid,particles,occupied_cells))
            {
                (*valid)(p)=false;
                continue;
            }
                
            if(debug_streamlines)
                Add_Text(particles.X(p),LOG::sprintf("%d",p),VECTOR<T,3>());

            for(int i=0;i<substeps;i++)
            {
                TV X_f=X_forward(p);
                TV X_b=X_backward(p);

                if(valid_forward(p)){
                    X_f=Advance_Particle(grid,X_forward(p),3,dt,*std::get<2>(name(fs)),1,*std::get<0>(name(fs)),valid_forward(p),cyan,png_grid,png_array,colors(p));
                    occupied_cells(ogrid.Cell_I(X_f))=true;
                }
                if(valid_backward(p)){
                    X_b=Advance_Particle(grid,X_backward(p),3,dt,*std::get<2>(name(fs)),-1,*std::get<0>(name(fs)),valid_backward(p),red,png_grid,png_array,colors(p));
                    occupied_cells(ogrid.Cell_I(X_b))=true;
                }

                if(valid_forward(p) && valid_backward(p))
                {
                    SEGMENT_2D<T> segment_f(X_forward(p),X_f);
                    SEGMENT_2D<T> segment_b(X_backward(p),X_b);
                    T a,b;
                    TV f_b=segment_f.Shortest_Vector_Between_Segments(segment_b,a,b);
                    TV center_pt=X_forward(p)*(1-a) + X_f*a - .5 * f_b;

                    T step=max(segment_f.Size(),segment_b.Size());
                    if(debug_streamlines && f_b.Magnitude() < step) Add_Segment(X_forward(p)*(1-a) + X_f*a,X_backward(p)*(1-b) + X_b*b,green);
                    min_substeps=max(threshold/step,1.)*2;
                    if((i>min_substeps && f_b.Magnitude() < threshold))
                    {
                        if(debug_streamlines)
                        {
                            Add_Ball(center_pt,.1,green);
                            std::cout<<"streamline closed "<<p<<" after "<<i<<" substeps"<<std::endl;
                        }
                        break;
                    }
                }
                else if(!valid_forward(p) && !valid_backward(p))
                {
                    if(debug_streamlines) std::cout<<"domain exit after "<<i<<" substeps"<<std::endl;
                    break;
                }
                
                X_forward(p) = X_f;
                X_backward(p) = X_b;
            }
        }
        Flush_Frame(std::get<1>(name(fs)).c_str());
        Write_To_File(stream_type,viewer_dir.Frame_File("mpm_particles"),particles);
        std::string png_filename=single_streamline>0?"s":"streamlines";
        png_filename += LOG::sprintf("-%s.png",std::get<1>(name(fs)));
        Write_Png(png_filename,png_array,antialiasing,resolution,png_grid,single_streamline?particles.X(single_streamline):TV());
    }
}


int main(int argc,char* argv[])
{
    PARSE_ARGS parse_args(argc,argv);
    Run(parse_args);

    return 0;
}
