//#####################################################################
// Copyright 2021, Craig Schroeder, Tamar Shinar.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Arrays/ARRAYS_ND.h>
#include <Core/Arrays/FACE_ARRAYS.h>
#include <Core/Boundaries/BOUNDARY.h>
#include <Core/Geometry_Particles/DEBUG_PARTICLES.h>
#include <Core/Grids/CELL_ITERATOR.h>
#include <Core/Grids/FACE_INDEX.h>
#include <Core/Grids/FACE_RANGE_ITERATOR.h>
#include <Core/Grids/GRID.h>
#include <Core/Interpolation/CUBIC_INTERPOLATION_FACE.h>
#include <Core/Interpolation/CUBIC_SPLINE_INTERPOLATION_FACE.h>
#include <Core/Interpolation/DIV_FREE_BALSARA.h>
#include <Core/Interpolation/DIV_FREE_INTERPOLATION.h>
#include <Core/Interpolation/DIV_FREE_INTERPOLATION_C1.h>
#include <Core/Interpolation/DIV_FREE_INTERPOLATION_C1_INTERP.h>
#include <Core/Interpolation/DIV_FREE_INTERPOLATION_INTERP.h>
#include <Core/Interpolation/GENERAL_INTERPOLATION_FACE.h>
#include <Core/Interpolation/LINEAR_INTERPOLATION_FACE.h>
#include <Core/Parsing/PARSE_ARGS.h>
#include <Core/Random_Numbers/RANDOM_NUMBERS.h>
#include <Core/Seeding/POISSON_DISK.h>
#include <Core/Vectors/VECTOR.h>
#include "INTERPOLATION_SCHEMES.h"
#include "VELOCITY_FIELDS.h"
#include <fstream>
#include <x86intrin.h>

using namespace PhysBAM;

typedef float RW;
typedef float T;

template<int d>
void Run(PARSE_ARGS& parse_args)
{
    typedef VECTOR<T,d> TV;
    typedef VECTOR<int,TV::m> IV;
    std::string sim_dir;
    int velocity_field=0;
    bool plot_error=false;
    parse_args.Add("-sim",&sim_dir,"dir","simulation directory to trace particles on");
    parse_args.Add("-vel",&velocity_field,"velocity field","which velocity field to use (0-7)");
    parse_args.Add("-err",&plot_error,"plot errors instead of velocities");
    parse_args.Parse();
    
    GRID<TV> grid(IV()+8,RANGE<TV>::Unit_Box());
    ARRAY<T,FACE_INDEX<TV::m> > data(grid.Cell_Indices(4));
    IV Ai,Bi;
    Ai(0)=1;
    Ai(1)=2;
    if constexpr(TV::m==3) Ai(2)=1;
    Bi(0)=7;
    Bi(1)=4;
    if constexpr(TV::m==3) Bi(2)=5;

    TV A=grid.Face_X({0,Ai}),B=grid.Face_X({0,Bi});

    auto field=Velocity_Field<TV>(velocity_field);
    for(FACE_RANGE_ITERATOR<TV::m> it(data.domain);it.Valid();it.Next())
    {
        TV X=grid.Face_X(it.face);
        data(it.face)=field(X)(it.face.axis);
    }

    auto interps=Interpolation_Schemes<TV>();
    //df_c1_i.param=param_c1_int;
    int n=100000;

    IV prev_cell = grid.Cell_I(A);

    ARRAY<long> best(interps.m);
    for(auto&b: best) b=1<<30;
    for(int r=0;r<10;r++)
    for(int s=0;s<interps.m;s++)
    {
        auto interp=interps(s).interp;
        long t0 = __rdtsc();
        TV sum;
        for(int i=0;i<=n;i++)
        {
            T t=(T)i/n;
            TV X=A+t*(B-A);
            IV cell=grid.Cell_I(X);
            prev_cell = cell;
            sum += interp->Interpolate_Face(grid,data,X);
        }
        long t1 = __rdtsc();
        LOG::printf("raw %P    (%P)\n",(t1-t0)/3.2e6,interps(s).name);
        best(s)=min(best(s),t1-t0);
    }
    for(int i=0;i<interps.m;i++)
        LOG::printf("scheme %i %P  (%P)\n",i,best(i)/3.2e6,interps(i).name);
    for(int i=0;i<interps.m;i++)
        LOG::printf("scheme %i %P  (%P)\n",i,(double)best(i)/best(0),interps(i).name);
}

int main(int argc, char* argv[])
{
    PARSE_ARGS parse_args(argc,argv);
    int dim=2;
    parse_args.Add("-dim",&dim,"dim","dimension to test");
    parse_args.Parse(true);
    if(dim==2) Run<2>(parse_args);
    else Run<3>(parse_args);
    
    return 0;
}

